import dagster as dg
from packaging.version import Version

if Version(dg.__version__) < Version("1.11.6"):
    from dagster._core.remote_representation import CodeLocationOrigin, InProcessCodeLocationOrigin
else:
    from dagster._core.remote_origin import InProcessCodeLocationOrigin, CodeLocationOrigin  # pyright: ignore[reportMissingImports,reportUnusedImport]  # noqa: F401,I001

assert InProcessCodeLocationOrigin is not None
assert CodeLocationOrigin is not None


def get_saved_path(result: dg.ExecuteInProcessResult, asset_name: str) -> str:
    path = (
        list(filter(lambda evt: evt.is_handled_output, result.events_for_node(asset_name)))[0]
        .event_specific_data.metadata["path"]  # type: ignore
        .value
    )  # type: ignore[index,union-attr]
    assert isinstance(path, str)
    return path


__all__ = ["CodeLocationOrigin", "InProcessCodeLocationOrigin", "get_saved_path"]



import logging
import os
import sys
from collections.abc import Iterator

import dagster as dg
import pytest
from _pytest.tmpdir import TempPathFactory
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin

# this import only works under `pytest` because of pytest.ini_options.pythonpath setting
from tests.utils import InProcessCodeLocationOrigin  # pyright:ignore[reportAttributeAccessIssue]

logging.getLogger("alembic").setLevel(logging.WARNING)


@pytest.fixture
def dagster_instance(tmp_path_factory: TempPathFactory) -> dg.DagsterInstance:
    return dg.DagsterInstance.ephemeral(tempdir=str(tmp_path_factory.mktemp("dagster_home")))


@pytest.fixture
def local_ray_address() -> Iterator[str]:
    import ray

    context = ray.init(
        ignore_reinit_error=True, runtime_env={"env_vars": {"RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING": "1"}}
    )

    yield "auto"

    context.disconnect()


@pytest.fixture
def code_location_origin() -> InProcessCodeLocationOrigin:
    return InProcessCodeLocationOrigin(
        loadable_target_origin=LoadableTargetOrigin(
            executable_path=sys.executable,
            module_name=("dagster_ray.tests.kuberay.conftest"),
            working_directory=os.getcwd(),
            attribute="lmao",
        ),
        location_name="test_location",
    )



import socket
import time
from datetime import datetime, timedelta
from typing import Any, cast

import dagster as dg
import pytest
import ray  # noqa: TID253
from pytest_kubernetes.providers import AClusterManager
from tempenv import TemporaryEnvironment

from dagster_ray import Lifecycle, RayResource
from dagster_ray._base.cluster_sharing_lock import ClusterSharingLock
from dagster_ray.configs import DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_ENV_VAR
from dagster_ray.kuberay import (
    KubeRayCluster,
    KubeRayClusterClientResource,
    RayClusterConfig,
    cleanup_expired_kuberay_clusters,
)
from dagster_ray.kuberay.client import RayClusterClient
from dagster_ray.kuberay.configs import ClusterSharing, RayClusterSpec
from dagster_ray.kuberay.jobs import delete_kuberay_clusters
from dagster_ray.kuberay.ops import DeleteKubeRayClustersConfig, RayClusterRef
from tests.kuberay.conftest import KUBERNETES_CONTEXT
from tests.kuberay.utils import NAMESPACE, get_random_free_port
from tests.utils import CodeLocationOrigin  # pyright: ignore[reportAttributeAccessIssue]

TESTING_LABELS = {"dagster/code-location": "test_location"}


def delete_shared_rayclusters(raycluster_client: RayClusterClient | KubeRayClusterClientResource):
    raycluster_client = (
        raycluster_client
        if isinstance(raycluster_client, RayClusterClient)
        else RayClusterClient(kube_config=raycluster_client.kube_config, kube_context=raycluster_client.kube_context)
    )
    clusters = raycluster_client.list(namespace=NAMESPACE).get("items", [])
    for cluster in clusters:
        if cluster["metadata"].get("labels", {}).get("dagster/cluster-sharing") == "true":
            raycluster_client.delete(name=cluster["metadata"]["name"], namespace=cluster["metadata"]["namespace"])


@pytest.fixture(scope="session")
def raycluster_client_resource(k8s_with_kuberay: AClusterManager):
    return KubeRayClusterClientResource(kube_config=str(k8s_with_kuberay.kubeconfig), kube_context=KUBERNETES_CONTEXT)


@pytest.fixture(scope="session")
def raycluster_client(k8s_with_kuberay: AClusterManager):
    return RayClusterClient(kube_config=str(k8s_with_kuberay.kubeconfig), kube_context=KUBERNETES_CONTEXT)


@pytest.fixture
def shared_clusters_cleanup(raycluster_client: RayClusterClient):
    delete_shared_rayclusters(raycluster_client)
    yield
    delete_shared_rayclusters(raycluster_client)


def test_instantiate_defaults():
    _ = KubeRayCluster()


@pytest.fixture(scope="session")
def ray_cluster_resource(
    k8s_with_kuberay: AClusterManager,
    raycluster_client_resource: KubeRayClusterClientResource,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
) -> KubeRayCluster:
    redis_port = get_random_free_port()

    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(connect=False),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=redis_port,
    )


@pytest.fixture(scope="session")
def ray_cluster_resource_skip_cleanup(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    raycluster_client_resource: KubeRayClusterClientResource,
) -> KubeRayCluster:
    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(
            connect=False,
            cleanup="never",
        ),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
    )


@pytest.fixture(scope="session")
def ray_cluster_resource_skip_create(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    raycluster_client_resource: KubeRayClusterClientResource,
) -> KubeRayCluster:
    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(create=False),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
    )


@pytest.fixture(scope="session")
def ray_cluster_resource_skip_wait(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    raycluster_client_resource: KubeRayClusterClientResource,
) -> KubeRayCluster:
    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(wait=False),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
    )


@ray.remote
def get_hostname():
    return socket.gethostname()


def ensure_kuberay_cluster_correctness(
    ray_cluster: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
    context: dg.AssetExecutionContext,
):
    with k8s_with_kuberay.port_forwarding(
        target=f"svc/{ray_cluster.name}-head-svc",
        source_port=cast(int, ray_cluster.redis_port),
        target_port=10001,
        namespace=ray_cluster.namespace,
    ):
        # now we can access the head node
        # hack the _host attribute to point to the port-forwarded address
        ray_cluster._host = "127.0.0.1"
        ray_cluster.connect(context)  # normally this would happen automatically during resource setup
        assert ray_cluster.context is not None

        # make sure a @remote function runs inside the cluster
        # not in localhost
        assert ray_cluster.name in ray.get(get_hostname.remote())

        ray_cluster_description = ray_cluster.client.get(ray_cluster.name, namespace=ray_cluster.namespace)
        assert ray_cluster_description["metadata"]["labels"]["dagster/run-id"] == context.run_id


def test_kuberay_cluster_resource(
    ray_cluster_resource: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
    raycluster_client: RayClusterClient,
):
    @dg.asset
    # testing RayResource type annotation too!
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource) -> None:
        # port-forward to the head node
        # because it's not possible to access it otherwise

        assert isinstance(ray_cluster, KubeRayCluster)

        ensure_kuberay_cluster_correctness(
            ray_cluster,
            k8s_with_kuberay,
            context,
        )

    result = dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource},
    )

    # make sure the RayCluster is cleaned up

    assert (
        len(
            raycluster_client.list(
                namespace=ray_cluster_resource.namespace, label_selector=f"dagster/run-id={result.run_id}"
            )["items"]
        )
        == 0
    )


def test_kuberay_cluster_resource_skip_create(
    ray_cluster_resource_skip_create: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource) -> None:
        assert isinstance(ray_cluster, KubeRayCluster)

        # call create and wait manually
        ray_cluster.create(context)
        ray_cluster.wait(context)

        ensure_kuberay_cluster_correctness(
            ray_cluster,
            k8s_with_kuberay,
            context,
        )

    dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource_skip_create},
    )


def test_kuberay_cluster_resource_skip_wait(
    ray_cluster_resource_skip_wait: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource) -> None:
        assert isinstance(ray_cluster, KubeRayCluster)

        # call wait manually
        ray_cluster.wait(context)

        ensure_kuberay_cluster_correctness(
            ray_cluster,
            k8s_with_kuberay,
            context,
        )

    dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource_skip_wait},
    )


def test_ray_cluster_builder_debug():
    kuberay_cluster = KubeRayCluster(enable_debug_post_mortem=True, image="test")
    kuberay_cluster._cluster_name = "test-cluster"
    context = dg.build_init_resource_context()

    ray_cluster_config = kuberay_cluster.ray_cluster.to_k8s(
        context, env_vars=kuberay_cluster.get_env_vars_to_inject(), labels={"foo": "bar"}, image="test"
    )
    assert ray_cluster_config["metadata"]["labels"]["foo"] == "bar"
    for group_spec in [ray_cluster_config["spec"]["headGroupSpec"], *ray_cluster_config["spec"]["workerGroupSpecs"]]:
        for container in group_spec["template"]["spec"]["containers"]:
            assert {"name": "RAY_DEBUG_POST_MORTEM", "value": "1"} in container["env"], container

    kuberay_cluster = KubeRayCluster(enable_tracing=True)
    kuberay_cluster._cluster_name = "test-cluster"
    ray_cluster_config = kuberay_cluster.ray_cluster.to_k8s(
        context, env_vars=kuberay_cluster.get_env_vars_to_inject(), image="test"
    )
    for group_spec in [ray_cluster_config["spec"]["headGroupSpec"], *ray_cluster_config["spec"]["workerGroupSpecs"]]:
        for container in group_spec["template"]["spec"]["containers"]:
            assert {"name": "RAY_PROFILING", "value": "1"} in container["env"], container

    kuberay_cluster = KubeRayCluster(enable_actor_task_logging=True)
    kuberay_cluster._cluster_name = "test-cluster"
    ray_cluster_config = kuberay_cluster.ray_cluster.to_k8s(
        context, env_vars=kuberay_cluster.get_env_vars_to_inject(), image="test"
    )
    kuberay_cluster._cluster_name = "test-cluster"
    for group_spec in [ray_cluster_config["spec"]["headGroupSpec"], *ray_cluster_config["spec"]["workerGroupSpecs"]]:
        for container in group_spec["template"]["spec"]["containers"]:
            assert {"name": "RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING", "value": "1"} in container["env"], container


def test_cleanup_expired_kuberay_clusters_sensor_skip(
    code_location_origin: CodeLocationOrigin,
    dagster_instance: dg.DagsterInstance,
    raycluster_client: RayClusterClient,
    shared_clusters_cleanup,
):
    # run the sensor - it shouldn't request any runs
    context = dg.build_sensor_context(instance=dagster_instance)
    context._code_location_origin = code_location_origin
    for item in cleanup_expired_kuberay_clusters(  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
        context, raycluster_client=raycluster_client
    ):
        assert isinstance(item, dg.SkipReason)


# TODO: investigate whether this could be reduced without risking on flakiness
TESTING_CLUSTER_EXPIRATION_SECONDS = 10


def test_cleanup_expired_kuberay_clusters_sensor_cluster_sharing_request(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    code_location_origin: CodeLocationOrigin,
    dagster_instance: dg.DagsterInstance,
    raycluster_client_resource: KubeRayClusterClientResource,
    raycluster_client: RayClusterClient,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    shared_clusters_cleanup,
):
    # run the sensor - it shouldn't request any runs
    context = dg.build_sensor_context(instance=dagster_instance)
    context._code_location_origin = code_location_origin
    for item in cleanup_expired_kuberay_clusters(  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
        context, raycluster_client=raycluster_client
    ):
        assert isinstance(item, dg.SkipReason)

    ray_cluster = KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(wait=False),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            # dagster/code-location should be added manually since there is no (normal) way to set the code location name for `dg.materialize`
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
        cluster_sharing=ClusterSharing(enabled=True, ttl_seconds=TESTING_CLUSTER_EXPIRATION_SECONDS),
    )

    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource):
        return "Hello, World!"

    res = dg.materialize(assets=[my_asset], resources={"ray_cluster": ray_cluster})
    clusters = raycluster_client.list(
        namespace=ray_cluster.namespace,
        label_selector=f"dagster/run-id={res.run_id}",
    )["items"]

    # cluster should not be cleaned up yet since it has a TTL of 10 seconds
    assert len(clusters) == 1
    cluster = clusters[0]

    assert cluster["metadata"]["labels"].get("dagster/cluster-sharing") == "true"
    assert cluster["metadata"]["labels"].get("dagster/code-location") == "test_location"

    locks = ClusterSharingLock.parse_all_locks(cluster["metadata"]["annotations"])
    assert len(locks) == 1
    lock = locks[0]
    assert lock.ttl_seconds == TESTING_CLUSTER_EXPIRATION_SECONDS
    assert not lock.is_expired

    # wait until the lock expires
    time.sleep((lock.expired_at - datetime.now()).total_seconds() + 0.5)

    assert lock.is_expired

    # run the sensor - it should request a run
    context = dg.build_sensor_context(instance=dagster_instance)
    context._code_location_origin = code_location_origin
    for item in cleanup_expired_kuberay_clusters(  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
        context, raycluster_client=raycluster_client
    ):
        assert isinstance(item, dg.RunRequest)
        assert (
            item.run_config["ops"]["delete_kuberay_clusters_op"]["config"]["clusters"][0]["name"]
            == cluster["metadata"]["name"]
        )


def test_cleanup_expired_kuberay_clusters_sensor_no_cluster_sharing_request(
    code_location_origin: CodeLocationOrigin,
    dagster_instance: dg.DagsterInstance,
    raycluster_client: RayClusterClient,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    dagster_ray_image: str,
    raycluster_client_resource: KubeRayClusterClientResource,
):
    ray_cluster = KubeRayCluster(
        image=dagster_ray_image,
        lifecycle=Lifecycle(wait=False, cleanup="never"),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
    )

    with TemporaryEnvironment(
        {DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_ENV_VAR: str(int(TESTING_CLUSTER_EXPIRATION_SECONDS))}
    ):
        # run the sensor - it shouldn't request any runs
        context = dg.build_sensor_context(instance=dagster_instance)
        context._code_location_origin = code_location_origin
        for item in cleanup_expired_kuberay_clusters(  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
            context, raycluster_client=raycluster_client
        ):
            assert isinstance(item, dg.SkipReason)

        @dg.asset
        def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource):
            return "Hello, World!"

        res = dg.materialize(assets=[my_asset], resources={"ray_cluster": ray_cluster})
        clusters = raycluster_client.list(
            namespace=ray_cluster.namespace,
            label_selector=f"dagster/run-id={res.run_id}",
        )["items"]
        assert len(clusters) == 1
        cluster = clusters[0]

        # run the sensor - it shouldn't request any runs yet
        context = dg.build_sensor_context(instance=dagster_instance)
        context._code_location_origin = code_location_origin
        for item in cleanup_expired_kuberay_clusters(  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
            context, raycluster_client=raycluster_client
        ):
            assert isinstance(item, dg.SkipReason)

        time.sleep(TESTING_CLUSTER_EXPIRATION_SECONDS)

        # run the sensor - it should request a run
        context = dg.build_sensor_context(instance=dagster_instance)
        context._code_location_origin = code_location_origin
        for item in cleanup_expired_kuberay_clusters(  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
            context, raycluster_client=raycluster_client
        ):
            assert isinstance(item, dg.RunRequest)
            assert (
                item.run_config["ops"]["delete_kuberay_clusters_op"]["config"]["clusters"][0]["name"]
                == cluster["metadata"]["name"]
            )


def test_cluster_sharing(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    raycluster_client_resource: KubeRayClusterClientResource,
    raycluster_client: RayClusterClient,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    dagster_instance: dg.DagsterInstance,
    code_location_origin: CodeLocationOrigin,
    shared_clusters_cleanup,
):
    ray_cluster = KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(wait=False),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
        cluster_sharing=ClusterSharing(enabled=True),
    )

    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource):
        annotations = ray_cluster.get_sharing_lock_annotations(context)
        assert len(list(filter(lambda x: x.startswith(f"dagster/lock-{context.run_id}"), annotations.keys()))) == 1
        return "Hello, World!"

    res = dg.materialize(assets=[my_asset], resources={"ray_cluster": ray_cluster})

    # with cluster sharing enabled the cluster should not be cleaned up
    clusters = raycluster_client.list(
        namespace=ray_cluster.namespace,
        label_selector=f"dagster/run-id={res.run_id}",
    )["items"]
    assert len(clusters) == 1

    cluster = clusters[0]

    assert cluster["metadata"]["labels"].get("dagster/cluster-sharing") == "true"

    locks = ClusterSharingLock.parse_all_locks(cluster["metadata"]["annotations"])
    assert len(locks) == 1
    lock = locks[0]
    assert lock.run_id == res.run_id
    assert lock.created_at + timedelta(seconds=cast(float, lock.ttl_seconds)) > datetime.now()
    assert not lock.is_expired

    # run the sensor - it shouldn't request any runs

    context = dg.build_sensor_context(instance=dagster_instance)
    context._code_location_origin = code_location_origin

    for item in cleanup_expired_kuberay_clusters(context, raycluster_client=raycluster_client):  # pyright: ignore[reportGeneralTypeIssues,reportOptionalIterable]
        assert isinstance(item, dg.SkipReason)

    # run the asset again, it shouldn't create a new RayCluster this time

    res_2 = dg.materialize(assets=[my_asset], resources={"ray_cluster": ray_cluster})

    # the old cluster should still be around and it should contain the new run in one of the lock annotations
    updated_cluster = raycluster_client.list(
        namespace=ray_cluster.namespace,
        label_selector=f"dagster/run-id={res.run_id}",
    )["items"][0]

    assert cluster["metadata"]["labels"] == updated_cluster["metadata"]["labels"]
    locks = ClusterSharingLock.parse_all_locks(updated_cluster["metadata"]["annotations"])
    assert len(locks) == 2
    assert locks[0].run_id != locks[1].run_id

    # the new run should not have any clusters associated with it
    assert (
        len(
            raycluster_client.list(
                namespace=ray_cluster.namespace,
                label_selector=f"dagster/run-id={res_2.run_id}",
            )["items"]
        )
        == 0
    )


def test_cluster_sharing_cleanup_expired(
    dagster_ray_image: str,
    raycluster_client_resource: KubeRayClusterClientResource,
    raycluster_client: RayClusterClient,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    shared_clusters_cleanup,
):
    ray_cluster = KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(wait=False),
        client=raycluster_client_resource,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE, "labels": TESTING_LABELS},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=get_random_free_port(),
        cluster_sharing=ClusterSharing(enabled=True, ttl_seconds=0.1),
    )

    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource):
        time.sleep(0.1)
        return "Hello, World!"

    res = dg.materialize(assets=[my_asset], resources={"ray_cluster": ray_cluster})
    clusters = raycluster_client.list(
        namespace=ray_cluster.namespace,
        label_selector=f"dagster/run-id={res.run_id}",
    )["items"]

    # cluster should be cleaned up since it's only lock has already expired
    assert len(clusters) == 0


def test_delete_kuberay_clusters_job(
    ray_cluster_resource_skip_cleanup: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
    raycluster_client: RayClusterClient,
):
    @dg.asset
    def my_asset(ray_cluster: RayResource):
        return 42

    result = dg.materialize(
        assets=[my_asset],
        resources={
            "ray_cluster": ray_cluster_resource_skip_cleanup,
        },
    )

    clusters = raycluster_client.list(
        namespace=ray_cluster_resource_skip_cleanup.namespace,
        label_selector=f"dagster/run-id={result.run_id}",
    )["items"]

    assert len(clusters) == 1

    delete_kuberay_clusters.execute_in_process(
        resources={
            "raycluster_client": raycluster_client,
        },
        run_config=dg.RunConfig(
            ops={
                "delete_kuberay_clusters_op": DeleteKubeRayClustersConfig(
                    namespace=ray_cluster_resource_skip_cleanup.namespace,
                    clusters=[
                        RayClusterRef(
                            name=clusters[0]["metadata"]["name"],
                        )
                    ],
                )
            }
        ),
    )

    assert not raycluster_client.list(
        namespace=ray_cluster_resource_skip_cleanup.namespace, label_selector=f"dagster/run-id={result.run_id}"
    )["items"]



#!/usr/bin/env python3
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "packaging",
# ]
# ///
"""
Generate optimized test matrix for CI.

This script reduces the number of test jobs by ensuring each component version
is tested at least once without testing all possible combinations (full cartesian product).
"""

import json
import sys
from typing import NamedTuple

from packaging.version import Version


class VersionConfig(NamedTuple):
    """Version configuration without OS."""

    py: str
    ray: str
    dagster: str
    kuberay: str

    def is_valid(self) -> bool:
        """
        Check if this configuration is valid based on dependency constraints.

        Returns True if valid, False if the combination is incompatible.
        """
        # Python 3.12+ requires Ray >= 2.37.0
        if Version(self.py) >= Version("3.12") and Version(self.ray) < Version("2.37.0"):
            return False

        # Add more validation rules here as needed

        return True

    def with_os(self, os: str) -> "TestConfig":
        """Create a TestConfig by adding an OS to this VersionConfig."""
        return TestConfig(os=os, **self._asdict())


class TestConfig(NamedTuple):
    """Complete test configuration including OS."""

    os: str
    py: str
    ray: str
    dagster: str
    kuberay: str


# Component versions
OS_VERSIONS = ["Ubuntu"]
PYTHON_VERSIONS = ["3.12", "3.11", "3.10"]
RAY_VERSIONS = ["2.49.2", "2.46.0"]
DAGSTER_VERSIONS = ["1.11.13", "1.11.6", "1.11.1", "1.10.21"]
KUBERAY_VERSIONS = ["1.4.0", "1.3.0", "1.2.2"]

# Component registry for DRY iteration
COMPONENTS = {
    "os": OS_VERSIONS,
    "py": PYTHON_VERSIONS,
    "ray": RAY_VERSIONS,
    "dagster": DAGSTER_VERSIONS,
    "kuberay": KUBERAY_VERSIONS,
}


def get_extremes(versions: list[str]) -> tuple[str, str]:
    """Get both latest and oldest versions using proper version comparison."""
    sorted_versions = sorted(versions, key=Version)
    return sorted_versions[-1], sorted_versions[0]


def generate_matrix() -> list[TestConfig]:
    """
    Generate test matrix with smart heuristic ensuring each version appears at most once:

    1. Generate base configurations (OS-agnostic)
       - Latest everything
       - Each component version individually with latest of everything else
       - Oldest combination
    2. Apply each base configuration to all OSes
    3. Deduplicate via set
    """
    # Get version components (exclude OS)
    version_components = {k: v for k, v in COMPONENTS.items() if k != "os"}

    # Get latest and oldest for each component
    extremes = {name: get_extremes(versions) for name, versions in version_components.items()}
    latest_versions = {name: latest for name, (latest, _) in extremes.items()}
    oldest_versions = {name: oldest for name, (_, oldest) in extremes.items()}

    # Generate base configurations (without OS), validating each before adding
    base_configs: set[VersionConfig] = set()

    # 1. Latest everything (if valid)
    latest_config = VersionConfig(**latest_versions)
    if latest_config.is_valid():
        base_configs.add(latest_config)

    # Track which versions have been used
    used_versions: dict[str, set[str]] = {name: {latest_versions[name]} for name in version_components}

    # 2. Each non-latest version with latest of everything else (if valid)
    for component_name, component_versions in version_components.items():
        for version in component_versions:
            if version not in used_versions[component_name]:
                config_dict = latest_versions.copy()
                config_dict[component_name] = version
                # Validate before adding
                config = VersionConfig(**config_dict)
                if config.is_valid():
                    base_configs.add(config)
                    used_versions[component_name].add(version)

    # 3. Oldest combination (if valid)
    oldest_config = VersionConfig(**oldest_versions)
    if oldest_config.is_valid():
        base_configs.add(oldest_config)

    # Apply each base configuration to all OSes
    configs: set[TestConfig] = {version_config.with_os(os) for os in OS_VERSIONS for version_config in base_configs}

    matrix = sorted(configs)

    # Verify all component versions are covered
    for component_name, component_versions in version_components.items():
        used_in_matrix = {getattr(config, component_name) for config in matrix}
        assert set(component_versions).issubset(used_in_matrix), (
            f"Not all {component_name} versions are in the matrix. Missing: {set(component_versions) - used_in_matrix}"
        )

    return matrix


def main():
    dry_run = "--dry-run" in sys.argv

    matrix = generate_matrix()

    # Calculate total possible combinations
    total_combinations = (
        len(OS_VERSIONS) * len(PYTHON_VERSIONS) * len(RAY_VERSIONS) * len(DAGSTER_VERSIONS) * len(KUBERAY_VERSIONS)
    )

    if dry_run:
        print(f"Generated {len(matrix)} test configurations (down from {total_combinations}):\n")
        for i, config in enumerate(matrix, 1):
            print(
                f"{i:2d}. {config.os}, Python {config.py}, Ray {config.ray}, "
                f"Dagster {config.dagster}, KubeRay {config.kuberay}"
            )
        print(f"\nTotal jobs: {len(matrix)} (reduced from {total_combinations})")

        # Verify all versions are covered
        print("\nCoverage check:")
        # Map field names to display names
        field_info = [
            ("os", "OS versions", None),
            ("py", "Python versions", Version),
            ("ray", "Ray versions", Version),
            ("dagster", "Dagster versions", Version),
            ("kuberay", "KubeRay versions", Version),
        ]
        for field, label, sort_key in field_info:
            unique_values = set(getattr(c, field) for c in matrix)
            sorted_values = sorted(unique_values, key=sort_key) if sort_key else sorted(unique_values)
            print(f"  {label}: {sorted_values}")
    else:
        # Output JSON for GitHub Actions - convert namedtuples to dicts
        output = {"include": [config._asdict() for config in matrix]}
        print(json.dumps(output))


if __name__ == "__main__":
    main()



from typing import TypeAlias

import dagster as dg

OpOrAssetExecutionContext: TypeAlias = dg.OpExecutionContext | dg.AssetExecutionContext

AnyDagsterContext: TypeAlias = dg.OpExecutionContext | dg.AssetExecutionContext | dg.InitResourceContext



import dagster as dg
from pydantic import Field

from dagster_ray.kuberay.client.raycluster.client import RayClusterClient


class RayClusterRef(dg.Config):
    name: str
    namespace: str | None = None


class DeleteKubeRayClustersConfig(dg.Config):
    namespace: str | None = "ray"
    match_labels: dict[str, str] | None = Field(
        default=None,
        description="Label selector to filter RayCluster resources to delete. Is mutually exclusive with `clusters`.",
    )
    clusters: list[RayClusterRef] | None = Field(
        default=None,
        description="List of `RayCluster` resources to delete. Namespaces can be omitted if the top-level namespace field is specified. Is mutually exclusive with `match_labels`.",
    )


@dg.op(description="Deletes RayCluster resources")
def delete_kuberay_clusters_op(
    context: dg.OpExecutionContext,
    config: DeleteKubeRayClustersConfig,
    raycluster_client: dg.ResourceParam[RayClusterClient],
) -> None:
    if config.match_labels and config.clusters:
        raise ValueError("Only one of `match_labels` or `clusters` can be specified")

    elif config.match_labels is not None:
        assert config.namespace is not None, "Namespace must be specified when using match_labels"
        label_selector = ",".join([f"{key}={value}" for key, value in config.match_labels.items()])
        response = raycluster_client.list(namespace=config.namespace, label_selector=label_selector)
        if clusters := response.get("items", []):
            for cluster in clusters:
                raycluster_client.delete(name=cluster["metadata"]["name"], namespace=config.namespace)
            context.log.info(
                f"{len(clusters)} `RayCluster` resources matching label selector `{config.match_labels}` have been deleted!"
            )

    elif config.clusters is not None:
        if config.clusters:
            for cluster in config.clusters:
                namespace = cluster.namespace or config.namespace
                assert namespace is not None, (
                    f"Namespace must be specified either as a top-level field or on each cluster entry. {cluster.name} does not have a namespace specified"
                )
                raycluster_client.delete(name=cluster.name, namespace=namespace)
            context.log.info(f"{len(config.clusters)} `RayCluster` resources have been deleted!")
    else:
        raise ValueError("Either `match_labels` or `clusters` must be specified")
